import grp
import os
import pwd
import subprocess

import yaml

from charmhelpers import fetch
from charmhelpers.core import hookenv
from charmhelpers.core import host
from charmhelpers.payload import (
    archive,
    execd,
    )


hooks = hookenv.Hooks()
config = hookenv.config()

# Globals
CHARM_FILES_DIR = os.path.join(hookenv.charm_dir(), 'files')
REQUIRED_PACKAGES = [
    'python-virtualenv', 'python-dev', 'python-pygit2', 'git',
    # Unfortunately we need build-essential to compile some extensions,
    # notably Twisted.  Using wheels rather than a pip-cache might let us
    # avoid this in future.
    'build-essential',
    ]
BASE_DIR = config['base_dir']
CODE_DIR = os.path.join(BASE_DIR, 'code')
VENV_DIR = os.path.join(BASE_DIR, 'venv')
LOGS_DIR = os.path.join(BASE_DIR, 'logs')
DATA_DIR = os.path.join(BASE_DIR, 'data')

CODE_USER = config['code_user']
CODE_GROUP = config['code_group']
USER = config['user']
GROUP = config['group']


def get_ports():
    return [s.strip() for s in config['ports'].split(',')]  # no array type


# Environment
def make_srv_location():
    hookenv.log('Creating directories...')

    for dir in (BASE_DIR, CODE_DIR):
        host.mkdir(dir, owner=CODE_USER, group=CODE_GROUP, perms=0o755)
    for dir in (LOGS_DIR, DATA_DIR):
        host.mkdir(dir, owner=USER, group=GROUP, perms=0o755)


def execd_preinstall(service_name):
    execd.execd_preinstall()


def create_users(service_name):
    hookenv.log('Creating users and groups for %s.' % service_name)
    code_password = host.pwgen()
    host.adduser(CODE_USER, code_password)
    host.add_user_to_group(CODE_USER, CODE_GROUP)
    password = host.pwgen()
    host.adduser(USER, password)
    host.add_user_to_group(USER, GROUP)


def unpack_source(service_name):
    hookenv.log('Deploying source...')

    make_srv_location()

    # Copy source archive
    archive_path = os.path.join(BASE_DIR, 'turnip.tar.gz')

    with open(os.path.join(CHARM_FILES_DIR, 'turnip.tar.gz')) as file:
        host.write_file(archive_path, file.read(), perms=0o644)

    # Unpack source
    archive.extract_tarfile(archive_path, CODE_DIR)
    os.chown(
        CODE_DIR,
        pwd.getpwnam(CODE_USER).pw_uid, grp.getgrnam(CODE_GROUP).gr_gid)
    host.lchownr(CODE_DIR, CODE_USER, CODE_GROUP)


def install_packages(service_name):
    hookenv.log('Installing system packages...')
    fetch.configure_sources(update=True)
    fetch.apt_install(REQUIRED_PACKAGES, fatal=True)


def install_python_packages(service_name):
    hookenv.log('Installing Python dependencies...')
    pip_cache = os.path.join(CHARM_FILES_DIR, 'pip-cache')
    code_reqs = os.path.join(CODE_DIR, 'requirements.txt')
    deploy_reqs = os.path.join(hookenv.charm_dir(), 'deploy-requirements.txt')

    pip_bin = os.path.join(VENV_DIR, 'bin', 'pip')

    subprocess.call([
        'sudo', '-u', CODE_USER, 'virtualenv', '--system-site-packages',
        VENV_DIR])
    subprocess.check_call([
        'sudo', '-u', CODE_USER, pip_bin, 'install', '--no-index',
        '--find-links={}'.format(pip_cache), '-r', code_reqs,
        '-r', deploy_reqs])
    subprocess.check_call([
        'sudo', '-u', CODE_USER, pip_bin, 'install', '--no-deps',
        '-e', CODE_DIR])


def ensure_data_dir_writable(service_name):
    os.chown(DATA_DIR, pwd.getpwnam(USER).pw_uid, grp.getgrnam(GROUP).gr_gid)
    os.chmod(DATA_DIR, 0o755)


def publish_wsgi_relations(service_name):
    # Publish the wsgi-file relation so the gunicorn subordinate can
    # serve us. Other WSGI containers could be made to work, as the most
    # gunicorn-specific thing is the --paste hack.
    config = hookenv.config()
    env = {
        'PATH': '%s:$PATH' % os.path.join(VENV_DIR, 'bin'),
        'REPO_STORE': os.path.join(DATA_DIR, 'repos'),
        'TURNIP_LOG_DIR': LOGS_DIR,
        'VIRTINFO_ENDPOINT': config['virtinfo_endpoint'],
        'AUTHENTICATION_ENDPOINT': config['authentication_endpoint'],
        }
    for relid in hookenv.relation_ids('wsgi-file'):
        hookenv.relation_set(
            relid,
            working_dir=CODE_DIR,
            wsgi_wsgi_file='--paste api.ini',  # XXX: Gross.
            wsgi_user=config['user'],
            wsgi_group=config['group'],
            port=config['port_api'],
            env_extra=' '.join('%s=%s' % (k, v) for (k, v) in env.items()))


def publish_website_relations(service_name):
    # Due to https://launchpad.net/bugs/1409763, this is an action rather
    # than a provided_data item.
    #
    # haproxy:reverseproxy can only relate to http interfaces.  Rather than
    # providing things like git-service-smart-ssh on nonsensical http
    # interfaces, we instead provide an artificial "website" relation that
    # tells haproxy about everything.
    config = hookenv.config()
    server_name = hookenv.local_unit().replace('/', '-')
    server_ip = str(hookenv.unit_private_ip())
    tcp_options = ['mode tcp', 'balance leastconn', 'option tcplog']
    http_options = ['mode http', 'balance leastconn', 'option httplog']
    haproxy_services = yaml.dump([
        {
            'service_name': 'git-service-git',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_pack_frontend']),
            'service_options': tcp_options,
            'servers': [[
                server_name, server_ip, str(config['port_pack_frontend']),
                'check',
                ]],
            },
        {
            'service_name': 'git-service-smart-http',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_smart_http']),
            'service_options': http_options + ['option httpchk'],
            'servers': [[
                server_name, server_ip, str(config['port_smart_http']),
                'check',
                ]],
            },
        {
            'service_name': 'git-service-smart-ssh',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_smart_ssh']),
            'service_options': tcp_options,
            'servers': [[
                server_name, server_ip, str(config['port_smart_ssh']),
                'check',
                ]],
            },
        {
            'service_name': 'turnip-api',
            'service_host': '0.0.0.0',
            'service_port': str(config['port_api']),
            'service_options': http_options + ['option httpchk /repo'],
            'servers': [[
                server_name, server_ip, str(config['port_api']),
                'check',
                ]],
            },
        ])
    for relid in hookenv.relation_ids('website'):
        hookenv.relation_set(
            relid,
            hostname=hookenv.unit_private_ip(),
            port=config['port_smart_http'],
            services=haproxy_services)
